from robot.api import logger

from .robotapp import get_robot_instance


def get_remote_url():
    """Get selenium URL for connecting to remote WebDriver."""
    se = get_robot_instance().get_library_instance('Selenium2Library')
    url = se._current_browser().command_executor._url

    return url


def get_session_id():
    """Get selenium browser session id."""
    se = get_robot_instance().get_library_instance('Selenium2Library')
    job_id = se._current_browser().session_id

    return job_id


def get_webdriver_remote():
    """Print the way connecting to remote selenium server."""
    remote_url = get_remote_url()
    session_id = get_session_id()

    code = 'from selenium import webdriver;' \
        'd=webdriver.Remote(command_executor="%s",' \
        'desired_capabilities={});' \
        'd.session_id="%s"' % (
            remote_url,
            session_id,
        )

    logger.console('''
DEBUG FROM CONSOLE
# geckodriver user please check https://stackoverflow.com/a/37968826/150841
%s
''' % (code))
    logger.info(code)

    return code



from robot.libdocpkg.model import LibraryDoc
from robot.libdocpkg.robotbuilder import KeywordDocBuilder, LibraryDocBuilder
from robot.libraries import STDLIBS
from robot.running.namespace import IMPORTER


def get_builtin_libs():
    """Get robotframework builtin library names."""
    return list(STDLIBS)


def get_libs():
    """Get imported robotframework library names."""
    return sorted(IMPORTER._library_cache._items, key=lambda _: _.name)


def get_libs_dict():
    """Get imported robotframework libraries as a name -> lib dict"""
    return {lib.name: lib for lib in IMPORTER._library_cache._items}


def match_libs(name=''):
    """Find libraries by prefix of library name, default all"""
    libs = [_.name for _ in get_libs()]
    matched = [_ for _ in libs if _.lower().startswith(name.lower())]
    return matched


class ImportedLibraryDocBuilder(LibraryDocBuilder):

    def build(self, lib):
        libdoc = LibraryDoc(
            name=lib.name,
            doc=self._get_doc(lib),
            doc_format=lib.doc_format,
        )
        libdoc.inits = self._get_initializers(lib)
        libdoc.keywords = KeywordDocBuilder().build_keywords(lib)
        return libdoc



class SingletonContext:
    in_step_mode = False
    current_runner = None
    current_runner_step = None
    current_source_path = ''
    current_source_lineno = 0
    last_command = ''

    def __new__(cls):
        if not hasattr(cls, 'instance'):
            cls.instance = super(SingletonContext, cls).__new__(cls)
        return cls.instance


context = SingletonContext()



import cmd
import os

from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import CompleteStyle, prompt


class BaseCmd(cmd.Cmd):
    """Basic REPL tool."""
    prompt = '> '
    repeat_last_nonempty_command = False

    def emptyline(self):
        """Do not repeat the last command if input empty unless forced to."""
        if self.repeat_last_nonempty_command:
            return super(BaseCmd, self).emptyline()

    def do_exit(self, arg):
        """Exit the interpreter. You can also use the Ctrl-D shortcut."""

        return True

    do_EOF = do_exit

    def help_help(self):
        """Help of Help command"""

        print('Show help message.')

    def do_pdb(self, arg):
        """Enter the python debuger pdb. For development only."""
        print('break into python debugger: pdb')
        import pdb
        pdb.set_trace()

    def get_cmd_names(self):
        """Get all command names of CMD shell."""
        pre = 'do_'
        cut = len(pre)
        return [_[cut:] for _ in self.get_names() if _.startswith(pre)]

    def get_help_string(self, command_name):
        """Get help document of command."""
        func = getattr(self, 'do_{0}'.format(command_name), None)
        if not func:
            return ''
        return func.__doc__

    def get_helps(self):
        """Get all help documents of commands."""
        return [(name, self.get_help_string(name) or name)
                for name in self.get_cmd_names()]

    def get_completer(self):
        """Get completer instance."""

    def pre_loop_iter(self):
        """Excute before every loop iteration."""

    def _get_input(self):
        if self.cmdqueue:
            return self.cmdqueue.pop(0)
        else:
            try:
                return self.get_input()
            except KeyboardInterrupt:
                return

    def loop_once(self):
        self.pre_loop_iter()
        line = self._get_input()
        if line is None:
            return

        if line == 'exit':
            line = 'EOF'

        line = self.precmd(line)
        if line == 'EOF':
            # do not run 'EOF' command to avoid override 'lastcmd'
            stop = True
        else:
            stop = self.onecmd(line)
        stop = self.postcmd(stop, line)
        return stop

    def cmdloop(self, intro=None):
        """Better command loop.

        override default cmdloop method
        """
        if intro is not None:
            self.intro = intro
        if self.intro:
            self.stdout.write(self.intro)
            self.stdout.write('\n')

        self.preloop()

        stop = None
        while not stop:
            stop = self.loop_once()

        self.postloop()

    def get_input(self):
        return input(prompt=self.prompt)


class PromptToolkitCmd(BaseCmd):
    """CMD shell using prompt-toolkit."""

    get_prompt_tokens = None
    prompt_style = None
    intro = '''\
Only accepted plain text format keyword separated with two or more spaces.
Type "help" for more information.\
'''

    def __init__(self, completekey='tab', stdin=None, stdout=None,
                 history_path=''):
        BaseCmd.__init__(self, completekey, stdin, stdout)
        self.history = FileHistory(os.path.expanduser(history_path))

    def get_input(self):
        kwargs = dict(
            history=self.history,
            auto_suggest=AutoSuggestFromHistory(),
            enable_history_search=True,
            completer=self.get_completer(),
            complete_style=CompleteStyle.MULTI_COLUMN,
        )
        if self.get_prompt_tokens:
            kwargs['style'] = self.prompt_style
            prompt_str = self.get_prompt_tokens(self.prompt)
        else:
            prompt_str = self.prompt
        try:
            line = prompt(message=prompt_str, **kwargs)
        except EOFError:
            line = 'EOF'
        return line



import re

from robot.libraries.BuiltIn import BuiltIn

from .robotlib import ImportedLibraryDocBuilder, get_libs
from .robotvar import assign_variable

try:
    from robot.variables.search import is_variable
except ImportError:
    from robot.variables import is_var as is_variable  # robotframework < 3.2

KEYWORD_SEP = re.compile('  +|\t')

_lib_keywords_cache = {}


def parse_keyword(command):
    """Split a robotframework keyword string."""
    # TODO use robotframework functions
    return KEYWORD_SEP.split(command)


def get_lib_keywords(library):
    """Get keywords of imported library."""
    if library.name in _lib_keywords_cache:
        return _lib_keywords_cache[library.name]

    lib = ImportedLibraryDocBuilder().build(library)
    keywords = []
    for keyword in lib.keywords:
        keywords.append({
            'name': keyword.name,
            'lib': library.name,
            'doc': keyword.doc,
            'summary': keyword.doc.split('\n')[0],
        })

    _lib_keywords_cache[library.name] = keywords
    return keywords


def get_keywords():
    """Get all keywords of libraries."""
    for lib in get_libs():
        yield from get_lib_keywords(lib)


def find_keyword(keyword_name):
    keyword_name = keyword_name.lower()
    return [keyword
            for lib in get_libs()
            for keyword in get_lib_keywords(lib)
            if keyword['name'].lower() == keyword_name]


def _execute_variable(robot_instance, variable_name, keyword, args):
    variable_only = not args
    if variable_only:
        display_value = ['Log to console', keyword]
        robot_instance.run_keyword(*display_value)
    else:
        variable_value = assign_variable(
            robot_instance,
            variable_name,
            args,
        )
        echo = '{0} = {1!r}'.format(variable_name, variable_value)
        return ('#', echo)


def run_keyword(robot_instance, keyword):
    """Run a keyword in robotframewrk environment."""
    if not keyword:
        return

    keyword_args = parse_keyword(keyword)
    keyword = keyword_args[0]
    args = keyword_args[1:]

    is_comment = keyword.strip().startswith('#')
    if is_comment:
        return

    variable_name = keyword.rstrip('= ')
    if is_variable(variable_name):
        return _execute_variable(robot_instance, variable_name, keyword, args)
    else:
        output = robot_instance.run_keyword(keyword, *args)
        if output:
            return ('<', repr(output))


def run_debug_if(condition, *args):
    """Runs DEBUG if condition is true."""

    return BuiltIn().run_keyword_if(condition,
                                    'DebugLibrary.DEBUG',
                                    *args)



from .keywords import DebugKeywords
from .version import VERSION

"""A debug library and REPL for RobotFramework."""


class DebugLibrary(DebugKeywords):
    """Debug Library for RobotFramework."""

    ROBOT_LIBRARY_SCOPE = 'GLOBAL'
    ROBOT_LIBRARY_VERSION = VERSION



import inspect

from .globals import context


class RobotLibraryStepListenerMixin:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self):
        super(RobotLibraryStepListenerMixin, self).__init__()
        self.ROBOT_LIBRARY_LISTENER = [self]

    def _start_keyword(self, name, attrs):
        context.current_source_path = ''
        context.current_source_lineno = 0

        if not is_step_mode():
            return

        find_runner_step()
        step = context.current_runner_step

        if hasattr(step, 'lineno'):
            path = step.source
            lineno = step.lineno
            lineno_0_based = lineno - 1
            context.current_source_path = path
            context.current_source_lineno = lineno
            print('> {}({})'.format(path, lineno))
            line = (open(path).readlines()[lineno_0_based].strip())
            print('-> {}'.format(line))

        if attrs['assign']:
            assign = '%s = ' % ', '.join(attrs['assign'])
        else:
            assign = ''
            name = '{}.{}'.format(attrs['libname'], attrs['kwname'])

        translated = '{}{}  {}'.format(assign, name, '  '.join(attrs['args']))
        print('=> {}'.format(translated))

        # callback debug interface
        self.debug()


def find_runner_step():
    stack = inspect.stack()
    for frame in stack:
        if frame.function == 'run_steps':
            arginfo = inspect.getargvalues(frame.frame)
            context.current_runner = arginfo.locals.get('runner')
            context.current_runner_step = arginfo.locals.get('step')


def set_step_mode(on=True):
    context.in_step_mode = on


def is_step_mode():
    return context.in_step_mode




from .robotkeyword import parse_keyword

SELENIUM_WEBDRIVERS = ['firefox', 'chrome', 'ie',
                       'opera', 'safari', 'phantomjs', 'remote']


def start_selenium_commands(arg):
    """Start a selenium webdriver and open url in browser you expect.

    arg:  [<url> or google]  [<browser> or firefox]
    """
    yield 'import library  SeleniumLibrary'

    # Set defaults, overriden if args set
    url = 'http://www.google.com/'
    browser = 'firefox'
    if arg:
        args = parse_keyword(arg)
        if len(args) == 2:
            url, browser = args
        else:
            url = arg
    if '://' not in url:
        url = 'http://' + url

    yield 'open browser  %s  %s' % (url, browser)



from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.styles import Style

NORMAL_STYLE = Style.from_dict({
    'head': 'fg:green',
    'message': 'fg:silver',
})

ERROR_STYLE = Style.from_dict({
    'head': 'fg:red',
    'message': 'fg:white',
})

DEBUG_PROMPT_STYLE = Style.from_dict({
    'prompt': 'blue',
})


def print_output(head, message, style=NORMAL_STYLE):
    """Print prompt-toolkit tokens to output."""
    tokens = FormattedText([
        ('class:head', '{0} '.format(head)),
        ('class:message', message),
        ('', ''),
    ])
    print_formatted_text(tokens, style=style)


def print_error(head, message, style=ERROR_STYLE):
    """Print to output with error style."""
    print_output(head, message, style=style)


def get_debug_prompt_tokens(prompt_text):
    """Print prompt-toolkit prompt."""
    return [
        ('class:prompt', prompt_text),
    ]



from prompt_toolkit.completion import Completer, Completion

from .robotkeyword import parse_keyword


class CmdCompleter(Completer):
    """Completer for debug shell."""

    def __init__(self, commands, cmd_repl=None):
        self.names = []
        self.displays = {}
        self.display_metas = {}
        for name, display, display_meta in commands:
            self.names.append(name)
            self.displays[name] = display
            self.display_metas[name] = display_meta
        self.cmd_repl = cmd_repl

    def _get_argument_completions(self, completer, document):
        """Using Cmd.py's completer to complete arguments."""
        end_idx = document.cursor_position_col
        line = document.current_line
        if line[:end_idx].rfind(' ') >= 0:
            begin_idx = line[:end_idx].rfind(' ') + 1
        else:
            begin_idx = 0
        prefix = line[begin_idx:end_idx]

        completions = completer(prefix, line, begin_idx, end_idx)
        for comp in completions:
            yield Completion(comp, begin_idx - end_idx, display=comp)

    def _get_custom_completions(self, cmd_name, document):
        completer = getattr(
            self.cmd_repl,
            'complete_{0}'.format(cmd_name),
            None,
        )
        if completer:
            yield from self._get_argument_completions(completer, document)

    def _get_command_completions(self, text):
        return (Completion(name,
                           -len(text),
                           display=self.displays.get(name, ''),
                           display_meta=self.display_metas.get(name, ''),
                           )
                for name in self.names
                if ((('.' not in name and '.' not in text)  # root level
                     or ('.' in name and '.' in text))  # library level
                    and name.lower().strip().startswith(text.strip()))
                )

    def get_completions(self, document, complete_event):
        """Compute suggestions."""
        text = document.text_before_cursor.lower()
        parts = parse_keyword(text)

        if len(parts) >= 2:
            cmd_name = parts[0].strip()
            yield from self._get_custom_completions(cmd_name, document)
        else:
            yield from self._get_command_completions(text)



from robot.version import get_version

ROBOT_VERION_RUNNER_GET_STEP_LINENO = '3.2'


class RobotNeedUpgrade(Exception):
    """Need upgrade robotframework."""


def check_version():
    if get_version() < ROBOT_VERION_RUNNER_GET_STEP_LINENO:
        raise RobotNeedUpgrade


def print_source_lines(source_file, lineno, before_and_after=5):
    check_version()

    if not source_file or not lineno:
        return

    lines = open(source_file).readlines()
    start_index = max(1, lineno - before_and_after - 1)
    end_index = min(len(lines) + 1, lineno + before_and_after)
    _print_lines(lines, start_index, end_index, lineno)


def print_test_case_lines(source_file, current_lineno):
    check_version()

    if not source_file or not current_lineno:
        return

    lines = open(source_file).readlines()

    # find the first line of current test case
    start_index = _find_first_lineno(lines, current_lineno)
    # find the last line of current test case
    end_index = _find_last_lineno(lines, current_lineno)

    _print_lines(lines, start_index, end_index, current_lineno)


def _find_last_lineno(lines, begin_lineno):
    line_index = begin_lineno - 1
    while line_index < len(lines):
        line = lines[line_index]
        if not _inside_test_case_block(line):
            break
        line_index += 1
    return line_index


def _find_first_lineno(lines, begin_lineno):
    line_index = begin_lineno - 1
    while line_index >= 0:
        line_index -= 1
        line = lines[line_index]
        if not _inside_test_case_block(line):
            break
    return line_index


def _inside_test_case_block(line):
    if line.startswith(' '):
        return True
    if line.startswith('\t'):
        return True
    if line.startswith('#'):
        return True
    return False


def _print_lines(lines, start_index, end_index, current_lineno):
    display_lines = lines[start_index:end_index]
    for lineno, line in enumerate(display_lines, start_index + 1):
        current_line_sign = ''
        if lineno == current_lineno:
            current_line_sign = '->'
        print('{:>3} {:2}\t{}'.format(lineno,
                                      current_line_sign,
                                      line.rstrip()))




from robot.api import logger
from robot.libraries.BuiltIn import BuiltIn
from robot.running.signalhandler import STOP_SIGNAL_MONITOR


def get_robot_instance():
    """Get robotframework builtin instance as context."""
    return BuiltIn()


def reset_robotframework_exception():
    """Resume RF after press ctrl+c during keyword running."""
    if STOP_SIGNAL_MONITOR._signal_count:
        STOP_SIGNAL_MONITOR._signal_count = 0
        STOP_SIGNAL_MONITOR._running_keyword = True
        logger.info('Reset last exception of DebugLibrary')



def assign_variable(robot_instance, variable_name, args):
    """Assign a robotframework variable."""
    variable_value = robot_instance.run_keyword(*args)
    robot_instance._variables.__setitem__(variable_name, variable_value)
    return variable_value



import os

from robot.api import logger
from robot.errors import ExecutionFailed, HandlerExecutionFailed

from .cmdcompleter import CmdCompleter
from .globals import context
from .prompttoolkitcmd import PromptToolkitCmd
from .robotapp import get_robot_instance, reset_robotframework_exception
from .robotkeyword import (get_keywords, get_lib_keywords, find_keyword,
                           run_keyword)
from .robotlib import get_builtin_libs, get_libs, get_libs_dict, match_libs
from .robotselenium import SELENIUM_WEBDRIVERS, start_selenium_commands
from .sourcelines import (RobotNeedUpgrade, print_source_lines,
                          print_test_case_lines)
from .steplistener import is_step_mode, set_step_mode
from .styles import (DEBUG_PROMPT_STYLE, get_debug_prompt_tokens, print_error,
                     print_output)

HISTORY_PATH = os.environ.get('RFDEBUG_HISTORY', '~/.rfdebug_history')


def run_robot_command(robot_instance, command):
    """Run command in robotframewrk environment."""
    if not command:
        return

    result = ''
    try:
        result = run_keyword(robot_instance, command)
    except HandlerExecutionFailed as exc:
        print_error('! keyword:', command)
        print_error('! handler execution failed:', exc.full_message)
    except ExecutionFailed as exc:
        print_error('! keyword:', command)
        print_error('! execution failed:', str(exc))
    except Exception as exc:
        print_error('! keyword:', command)
        print_error('! FAILED:', repr(exc))

    if result:
        head, message = result
        print_output(head, message)


class DebugCmd(PromptToolkitCmd):
    """Interactive debug shell for robotframework."""

    prompt_style = DEBUG_PROMPT_STYLE

    def __init__(self, completekey='tab', stdin=None, stdout=None):
        PromptToolkitCmd.__init__(self, completekey, stdin, stdout,
                                  history_path=HISTORY_PATH)
        self.robot = get_robot_instance()

    def get_prompt_tokens(self, prompt_text):
        return get_debug_prompt_tokens(prompt_text)

    def postcmd(self, stop, line):
        """Run after a command."""
        return stop

    def pre_loop_iter(self):
        """Reset robotframework before every loop iteration."""
        reset_robotframework_exception()

    def do_help(self, arg):
        """Show help message."""
        if not arg.strip():
            print('''\
Input Robotframework keywords, or commands listed below.
Use "libs" or "l" to see available libraries,
use "keywords" or "k" see the list of library keywords,
use the TAB keyboard key to autocomplete keywords.
Access https://github.com/xyb/robotframework-debuglibrary for more details.\
''')

        PromptToolkitCmd.do_help(self, arg)

    def get_completer(self):
        """Get completer instance specified for robotframework."""
        # commands
        commands = [(cmd_name, cmd_name, 'DEBUG command: {0}'.format(doc))
                    for cmd_name, doc in self.get_helps()]

        # libraries
        for lib in get_libs():
            commands.append((
                lib.name,
                lib.name,
                'Library: {0} {1}'.format(lib.name, lib.version),
            ))

        # keywords
        for keyword in get_keywords():
            # name with library
            name = '{0}.{1}'.format(keyword['lib'], keyword['name'])
            commands.append((
                name,
                keyword['name'],
                'Keyword: {0}'.format(keyword['summary']),
            ))
            # name without library
            commands.append((
                keyword['name'],
                keyword['name'],
                'Keyword[{0}.]: {1}'.format(keyword['lib'],
                                            keyword['summary']),
            ))

        return CmdCompleter(commands, self)

    def do_selenium(self, arg):
        """Start a selenium webdriver and open url in browser you expect.

        selenium  [<url>]  [<browser>]

        default url is google.com, default browser is firefox.
        """

        for command in start_selenium_commands(arg):
            print_output('#', command)
            run_robot_command(self.robot, command)

    def complete_selenium(self, text, line, begin_idx, end_idx):
        """Complete selenium command."""
        if len(line.split()) == 3:
            command, url, driver_name = line.lower().split()
            return [driver for driver in SELENIUM_WEBDRIVERS
                    if driver.startswith(driver_name)]
        elif len(line.split()) == 2 and line.endswith(' '):
            return SELENIUM_WEBDRIVERS
        return []

    complete_s = complete_selenium

    def default(self, line):
        """Run RobotFramework keywords."""
        command = line.strip()

        run_robot_command(self.robot, command)

    def _print_lib_info(self, lib, with_source_path=False):
        print_output('   {}'.format(lib.name), lib.version)
        if lib.doc:
            logger.console('       {}'.format(lib.doc.split('\n')[0]))
        if with_source_path:
            logger.console('       {}'.format(lib.source))

    def do_libs(self, args):
        """Print imported and builtin libraries, with source if `-s` specified.

        ls( libs ) [-s]
        """
        print_output('<', 'Imported libraries:')
        for lib in get_libs():
            self._print_lib_info(lib, with_source_path='-s' in args)
        print_output('<', 'Builtin libraries:')
        for name in sorted(get_builtin_libs()):
            print_output('   ' + name, '')

    do_ls = do_libs

    def complete_libs(self, text, line, begin_idx, end_idx):
        """Complete libs command."""
        if len(line.split()) == 1 and line.endswith(' '):
            return ['-s']
        return []

    complete_l = complete_libs

    def do_keywords(self, args):
        """Print keywords of libraries, all or starts with <lib_name>.

        k(eywords) [<lib_name>]
        """
        lib_name = args
        matched = match_libs(lib_name)
        if not matched:
            print_error('< not found library', lib_name)
            return
        libs = get_libs_dict()
        for name in matched:
            lib = libs[name]
            print_output('< Keywords of library', name)
            for keyword in get_lib_keywords(lib):
                print_output('   {}\t'.format(keyword['name']),
                             keyword['summary'])

    do_k = do_keywords

    def complete_keywords(self, text, line, begin_idx, end_idx):
        """Complete keywords command."""
        if len(line.split()) == 2:
            command, lib_name = line.split()
            return match_libs(lib_name)
        elif len(line.split()) == 1 and line.endswith(' '):
            return [_.name for _ in get_libs()]
        return []

    complete_k = complete_keywords

    def do_docs(self, keyword_name):
        """Get keyword documentation for individual keywords.

         d(ocs) [<keyword_name>]
        """

        keywords = find_keyword(keyword_name)
        if not keywords:
            print_error('< not find keyword', keyword_name)
        elif len(keywords) == 1:
            logger.console(keywords[0]['doc'])
        else:
            print_error('< found {} keywords'.format(len(keywords)),
                        ', '.join(keywords))

    do_d = do_docs

    def emptyline(self):
        """Repeat last nonempty command if in step mode."""
        self.repeat_last_nonempty_command = is_step_mode()
        return super(DebugCmd, self).emptyline()

    def append_command(self, command):
        """Append a command to queue."""
        self.cmdqueue.append(command)

    def append_exit(self):
        """Append exit command to queue."""
        self.append_command('exit')

    def do_step(self, args):
        """Execute the current line, stop at the first possible occasion."""
        set_step_mode(on=True)
        self.append_exit()  # pass control back to robot runner

    do_s = do_step

    def do_next(self, args):
        """Continue execution until the next line is reached or it returns."""
        self.do_step(args)

    do_n = do_next

    def do_continue(self, args):
        """Continue execution."""
        self.do_exit(args)

    do_c = do_continue

    def do_list(self, args):
        """List source code for the current file."""

        self.list_source(longlist=False)

    do_l = do_list

    def do_longlist(self, args):
        """List the whole source code for the current test case."""

        self.list_source(longlist=True)

    do_ll = do_longlist

    def list_source(self, longlist=False):
        """List source code."""
        if not is_step_mode():
            print('Please run `step` or `next` command first.')
            return

        if longlist:
            print_function = print_test_case_lines
        else:
            print_function = print_source_lines

        try:
            print_function(context.current_source_path,
                           context.current_source_lineno)
        except RobotNeedUpgrade:
            print('Please upgrade robotframework to support list source code:')
            print('    pip install "robotframework>=3.2" -U')

    def do_exit(self, args):
        """Exit debug shell."""
        set_step_mode(on=False)  # explicitly exit REPL will disable step mode
        self.append_exit()
        return super(DebugCmd, self).do_exit(args)

    def onecmd(self, line):
        # restore last command acrossing different Cmd instances
        self.lastcmd = context.last_command
        stop = super(DebugCmd, self).onecmd(line)
        context.last_command = self.lastcmd
        return stop




from functools import wraps


def memoize(function):
    """Memoization decorator"""
    memo = {}

    @wraps(function)
    def wrapper(*args, **kwargs):
        key = (args, frozenset(sorted(kwargs.items())))

        if key in memo:
            return memo[key]

        rv = function(*args, **kwargs)
        memo[key] = rv
        return rv
    return wrapper



#!/usr/bin/env python

import io
import os
import re

from setuptools import setup

ROOT = os.path.abspath(os.path.dirname(__file__))


def read(*names, **kwargs):
    with io.open(
        os.path.join(ROOT, *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^VERSION = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='robotframework-debuglibrary',
    version=find_version('DebugLibrary/version.py'),
    description='RobotFramework debug library and an interactive shell',
    long_description=read('README.rst'),
    author='Xie Yanbo',
    author_email='xieyanbo@gmail.com',
    license='New BSD',
    packages=['DebugLibrary'],
    entry_points={
        'console_scripts': [
            'rfshell = DebugLibrary.shell:shell',
            'rfdebug = DebugLibrary.shell:shell',
        ],
    },
    zip_safe=False,
    url='https://github.com/xyb/robotframework-debuglibrary/',
    keywords='robotframework,debug,shell,repl',
    install_requires=[
        'prompt-toolkit >= 2, < 3',  # 3.0 is not compatible with py3.5
        'robotframework >= 3.0',
    ],
    tests_require=['pexpect', 'coverage'],
    test_suite='tests.test_debuglibrary.suite',
    platforms=['Linux', 'Unix', 'Windows', 'MacOS X'],
    classifiers=[
        'Environment :: Console',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Utilities',
    ],
)



import os
import sys
import tempfile

from robot import run_cli

TEST_SUITE = b'''*** Settings ***
Library  DebugLibrary

** test case **
RFDEBUG REPL
    debug
'''


def shell():
    """A standalone robotframework shell."""

    default_no_logs = '-l None -x None -o None -L None -r None'

    with tempfile.NamedTemporaryFile(prefix='robot-debug-',
                                     suffix='.robot',
                                     delete=False) as test_file:
        test_file.write(TEST_SUITE)
        test_file.flush()

        if len(sys.argv) > 1:
            args = sys.argv[1:] + [test_file.name]
        else:
            args = default_no_logs.split() + [test_file.name]

        try:
            sys.exit(run_cli(args))
        finally:
            test_file.close()
            # pybot will raise PermissionError on Windows NT or later
            # if NamedTemporaryFile called with `delete=True`,
            # deleting test file seperated will be OK.
            if os.path.exists(test_file.name):
                os.unlink(test_file.name)


if __name__ == "__main__":
    # Usage: python -m DebugLibrary.shell
    shell()



import sys

from robot.libraries.BuiltIn import run_keyword_variant

from .debugcmd import DebugCmd
from .robotkeyword import run_debug_if
from .steplistener import RobotLibraryStepListenerMixin, is_step_mode
from .styles import print_output
from .webdriver import get_remote_url, get_session_id, get_webdriver_remote


class DebugKeywords(RobotLibraryStepListenerMixin):
    """Debug Keywords for RobotFramework."""

    def debug(self):
        """Open a interactive shell, run any RobotFramework keywords.

        Keywords separated by two space or one tab, and Ctrl-D to exit.
        """
        # re-wire stdout so that we can use the cmd module and have readline
        # support
        old_stdout = sys.stdout
        sys.stdout = sys.__stdout__

        show_intro = not is_step_mode()
        if show_intro:
            print_output('\n>>>>>', 'Enter interactive shell')

        self.debug_cmd = DebugCmd()
        if show_intro:
            self.debug_cmd.cmdloop()
        else:
            self.debug_cmd.cmdloop(intro='')

        show_intro = not is_step_mode()
        if show_intro:
            print_output('\n>>>>>', 'Exit shell.')

        # put stdout back where it was
        sys.stdout = old_stdout

    @run_keyword_variant(resolve=1)
    def debug_if(self, condition, *args):
        """Runs the Debug keyword if condition is true."""
        return run_debug_if(condition, *args)

    def get_remote_url(self):
        """Get selenium URL for connecting to remote WebDriver."""
        return get_remote_url()

    def get_session_id(self):
        """Get selenium browser session id."""
        return get_session_id()

    def get_webdriver_remote(self):
        """Print the way connecting to remote selenium server."""
        return get_webdriver_remote()
